@@ -6,14 +6,14 @@ use warnings;
 use Carp;
 use Lingua::KO::Hangul::Util;
 
-our $VERSION = '0.06';
+our $VERSION = '0.07';
 our $PACKAGE = __PACKAGE__;
 
 require Exporter;
 our @ISA = qw(Exporter);
 our @EXPORT = qw( NFC NFD NFKC NFKD );
-our @EXPORT_OK = qw( normalize getCanon getCompat
-		getComposite getCombinClass getExclusion);
+our @EXPORT_OK = qw( normalize decompose reorder compose 
+	getCanon getCompat getComposite getCombinClass getExclusion);
 our %EXPORT_TAGS = ( all => [ @EXPORT, @EXPORT_OK ] );
 
 our $Combin = do "unicore/CombiningClass.pl"
@@ -90,6 +90,20 @@ while($Decomp =~ /(.+)/g)
   }
 }
 
+sub getCanonList {
+  my @src = @_;
+  my @dec = map $Canon{$_} ? @{ $Canon{$_} } : $_, @src;
+  join(" ",@src) eq join(" ",@dec) ? @dec : getCanonList(@dec);
+  # condition @src == @dec is not ok.
+}
+
+sub getCompatList {
+  my @src = @_;
+  my @dec = map $Compat{$_} ? @{ $Compat{$_} } : $_, @src;
+  join(" ",@src) eq join(" ",@dec) ? @dec : getCompatList(@dec);
+  # condition @src == @dec is not ok.
+}
+
 foreach my $key (keys %Canon)  # exhaustive decomposition
 {
    $Canon{$key}  = [ getCanonList($key) ];
@@ -100,29 +114,160 @@ foreach my $key (keys %Compat) # exhaustive decomposition
    $Compat{$key} = [ getCompatList($key) ];
 }
 
-sub getCanonList
-{
-  my @src = @_;
-  my @dec = map $Canon{$_} ? @{ $Canon{$_} } : $_, @src;
-  join(" ",@src) eq join(" ",@dec) ? @dec : getCanonList(@dec);
-  # condition @src == @dec is not ok.
+##
+## Hangul Syllables
+##
+use constant Hangul_SBase  => 0xAC00;
+use constant Hangul_SFinal => 0xD7A3;
+use constant Hangul_SCount =>  11172;
+
+use constant Hangul_LBase  => 0x1100;
+use constant Hangul_LFinal => 0x1112;
+use constant Hangul_LCount =>     19;
+
+use constant Hangul_VBase  => 0x1161;
+use constant Hangul_VFinal => 0x1175;
+use constant Hangul_VCount =>     21;
+
+use constant Hangul_TBase  => 0x11A7;
+use constant Hangul_TFinal => 0x11C2;
+use constant Hangul_TCount =>     28;
+
+sub Hangul_IsS  ($) { Hangul_SBase <= $_[0] && $_[0] <= Hangul_SFinal }
+
+sub Hangul_IsL  ($) { Hangul_LBase <= $_[0] && $_[0] <= Hangul_LFinal }
+
+sub Hangul_IsV  ($) { Hangul_VBase <= $_[0] && $_[0] <= Hangul_VFinal }
+
+sub Hangul_IsT  ($) { Hangul_TBase <= $_[0] && $_[0] <= Hangul_TFinal }
+
+sub Hangul_IsLV ($) {
+    Hangul_SBase <= $_[0] && $_[0] <= Hangul_SFinal
+    && (($_[0] - Hangul_SBase ) % Hangul_TCount) == 0
 }
 
-sub getCompatList
-{
-  my @src = @_;
-  my @dec = map $Compat{$_} ? @{ $Compat{$_} } : $_, @src;
-  join(" ",@src) eq join(" ",@dec) ? @dec : getCompatList(@dec);
-  # condition @src == @dec is not ok.
+sub getCombinClass ($) { $Combin{$_[0]} || 0 }
+
+sub getCanon ($) { 
+   return exists $Canon{$_[0]}
+        ? pack('U*', @{ $Canon{$_[0]} })
+        : scalar decomposeHangul($_[0]);
 }
 
-sub NFD($){ _decompose(shift, 0) }
+sub getCompat ($) {
+    return exists $Compat{$_[0]}
+        ? pack('U*', @{ $Compat{$_[0]} })
+        : scalar decomposeHangul($_[0]);
+}
 
-sub NFKD($){ _decompose(shift, 1) }
+sub getComposite ($$) {
+    if(Hangul_IsL($_[0]) && Hangul_IsV($_[1])) {
+	my $lindex = $_[0] - Hangul_LBase;
+	my $vindex = $_[1] - Hangul_VBase;
+	return
+	(Hangul_SBase + ($lindex * Hangul_VCount + $vindex) * Hangul_TCount);
+    }
+    if(Hangul_IsLV($_[0]) && Hangul_IsT($_[1])) {
+	return($_[0] + $_[1] - Hangul_TBase);
+    }
+    return $Compos{ pack('U*', @_[0,1]) } || 0;
+}
 
-sub NFC($){ _compose(NFD(shift)) }
+sub getExclusion ($) {
+    return exists $Exclus{$_[0]} ? 1 : 0;
+}
 
-sub NFKC($){ _compose(NFKD(shift)) }
+##
+## string decompose(string, compat?)
+##
+sub decompose {
+    my $hash = $_[1] ? \%Compat : \%Canon;
+    return pack 'U*', map {
+	$hash->{ $_ } ? @{ $hash->{ $_ } } :
+	Hangul_IsS($_) ? decomposeHangul($_) : $_
+    } unpack('U*', $_[0]);
+}
+
+##
+## string reorder(string)
+##
+sub reorder {
+    my @src = unpack('U*', $_[0]);
+
+    for(my $i=0; $i < @src;){
+	$i++, next unless $Combin{ $src[$i] } 
+	    && $i+1 < @src && $Combin{ $src[$i+1] };
+	my $ini = $i;
+	$i++ while $i < @src && $Combin{ $src[$i] };
+        my @tmp = sort {
+		$Combin{ $src[$a] } <=> $Combin{ $src[$b] } || $a <=> $b
+	    } $ini .. $i - 1;
+
+	@src[ $ini .. $i - 1 ] = @src[ @tmp ];
+    }
+    return pack('U*', @src);
+}
+
+
+##
+## string compose(string)
+##
+## S : starter; NS : not starter;
+##
+## composable sequence begins at S.
+## S + S or (S + S) + S may be composed.
+## NS + NS must not be composed.
+##
+sub compose
+{
+    my @src = unpack('U*', $_[0]);
+
+    for(my $s = 0; $s+1 < @src; $s++){
+	next unless defined $src[$s] && ! $Combin{ $src[$s] }; # S only
+#	my($c, $blocked);
+	my($c, $blocked, $uncomposed_cc);
+	for(my $j = $s+1; $j < @src && !$blocked; $j++){
+	    ($Combin{ $src[$j] } ? $uncomposed_cc : $blocked) = 1;
+
+	    # S + C + S => S-S + C would be blocked. XXX 
+	    next if $blocked && $uncomposed_cc;
+
+	    # blocked by same CC
+	    next if defined $src[$j-1]   && $Combin{ $src[$j-1] } 
+		&& $Combin{ $src[$j-1] } == $Combin{ $src[$j] };
+
+	    $c = getComposite($src[$s], $src[$j]);
+
+	    # no composite or is exclusion
+	    next if !$c || $Exclus{$c};
+
+	    # replace by composite
+	    $src[$s] = $c; $src[$j] = undef;
+	    if($blocked) { $blocked = 0 } else { -- $uncomposed_cc }
+	}
+    }
+    return pack 'U*', grep defined(), @src;
+}
+
+##
+## "hhhh hhhh hhhh" to (dddd, dddd, dddd)
+##
+sub _getHexArray { map hex(), $_[0] =~ /([0-9A-Fa-f]+)/g }
+
+##
+## normalization forms
+##
+
+use constant CANON  => 0;
+use constant COMPAT => 1;
+
+sub NFD  ($) { reorder(decompose($_[0], CANON)) }
+
+sub NFKD ($) { reorder(decompose($_[0], COMPAT)) }
+
+sub NFC  ($) { compose(reorder(decompose($_[0], CANON))) }
+
+sub NFKC ($) { compose(reorder(decompose($_[0], COMPAT))) }
 
 sub normalize($$)
 {
@@ -134,86 +279,6 @@ sub normalize($$)
     croak $PACKAGE."::normalize: invalid form name: $form";
 }
 
-
-##
-## string _decompose(string, compat?)
-##
-sub _decompose
-{
-  my $str  = $_[0];
-  my $hash = $_[1] ? \%Compat : \%Canon;
-  my @ret;
-  my $retstr="";
-  foreach my $u (unpack 'U*', $str){
-    push @ret,
-      $hash->{ $u }  ? @{ $hash->{ $u } } :
-      _isHangul($u) ? decomposeHangul($u) : $u;
-  }
-  for(my $i=0; $i<@ret;){
-    $retstr .= pack('U', $ret[$i++]), next
-       unless $Combin{ $ret[$i] } && $i+1 < @ret && $Combin{ $ret[$i+1] };
-    my @tmp;
-    push(@tmp, $ret[$i++]) while $i < @ret && $Combin{ $ret[$i] };
-    $retstr .= pack 'U*', @tmp[
-      sort {
-        $Combin{ $tmp[$a] } <=> $Combin{ $tmp[$b] } || $a <=> $b
-      } 0 .. @tmp - 1,
-    ];
-  }
-  $retstr;
-}
-
-##
-## string _compose(string)
-##
-## S : starter; NS : not starter;
-##
-## composable sequence begins at S.
-## S + S or (S + S) + S may be composed.
-## NS + NS must not be composed.
-##
-sub _compose
-{
-  my @src = unpack('U*', composeHangul shift); # get codepoints
-  for(my $s = 0; $s+1 < @src; $s++){
-    next unless defined $src[$s] && ! $Combin{ $src[$s] }; # S only
-    my($c, $blocked);
-    for(my $j = $s+1; $j < @src && !$blocked; $j++){
-      $blocked = 1 if ! $Combin{ $src[$j] };
-
-      next if $j != $s + 1 && defined $src[$j-1]
-        && $Combin{ $src[$j-1] } && $Combin{ $src[$j] } 
-        && $Combin{ $src[$j-1] } == $Combin{ $src[$j] };
-
-      if(  # $c != 0, maybe.
-        $c = $Compos{pack('U*', @src[$s,$j])} and ! $Exclus{$c}
-      )
-      {
-        $src[$s] = $c; $src[$j] = undef; $blocked = 0;
-      }
-    }
-  }
-  pack 'U*', grep defined(), @src;
-}
-
-##
-## "hhhh hhhh hhhh" to (dddd, dddd, dddd)
-##
-sub _getHexArray
-{
-  my $str = shift;
-  map hex(), $str =~ /([0-9A-Fa-f]+)/g;
-}
-
-##
-## Hangul Syllables
-##
-sub _isHangul
-{
-  my $code = shift;
-  return 0xAC00 <= $code && $code <= 0xD7A3;
-}
-
 ##
 ## for Debug
 ##
@@ -223,26 +288,6 @@ sub _getCompat { wantarray ? %Compat : \%Compat }
 sub _getCompos { wantarray ? %Compos : \%Compos }
 sub _getExclus { wantarray ? %Exclus : \%Exclus }
 
-sub getCombinClass ($) {
-    return $Combin{$_[0]} || 0;
-}
-
-sub getCanon ($) {
-    return exists $Canon{$_[0]} ? pack('U*', @{ $Canon{$_[0]} }) : undef;
-}
-
-sub getCompat ($) {
-    return exists $Compat{$_[0]} ? pack('U*', @{ $Compat{$_[0]} }) : undef;
-}
-
-sub getComposite ($$) {
-    return $Compos{ pack('U*', @_[0,1]) } || 0;
-}
-
-sub getExclusion ($) {
-    return exists $Exclus{$_[0]} ? 1 : 0;
-}
-
 1;
 __END__
 
